from __future__ import annotations

from collections import defaultdict
from typing import Any, Dict, Hashable, List, Optional, Sequence, Set, Tuple, Type, TypeVar

import colorful as cf
from jax.tree_util import register_pytree_node

from .annotations import PyTree
from .display import BatchDimensionIterator, display_class, display_generic, display_key_and_value

__all__: List[str] = []


try:
    import networkx as nx
except ImportError:
    pass
else:
    T = TypeVar('T', bound=nx.Graph)

    def register_graph_as_jax_pytree(cls: Type[T]) -> None:
        def tree_unflatten(hashed: Hashable, trees: Sequence[PyTree]) -> T:
            node_dicts, edge_dicts = trees

            if not isinstance(node_dicts, dict):
                raise TypeError
            if not isinstance(edge_dicts, dict):
                raise TypeError

            graph = cls()
            graph.add_nodes_from(node_dicts.items())
            graph.add_edges_from([(source, target, data)
                                  for (source, target), data in edge_dicts.items()])

            return graph

        def tree_flatten(graph: T) -> Tuple[Sequence[PyTree], Hashable]:
            return ((dict(graph.nodes), dict(graph.edges)), None)

        register_pytree_node(cls, tree_flatten, tree_unflatten)

    register_graph_as_jax_pytree(nx.Graph)
    register_graph_as_jax_pytree(nx.DiGraph)

    try:
        from flax.serialization import (from_state_dict,  # type: ignore[import]
                                        register_serialization_state,
                                        to_state_dict)
    except ImportError:
        pass
    else:
        def register_graph_as_flax_state_dict(cls: Type[T]) -> None:
            def ty_to_state_dict(graph: T) -> Dict[str, Any]:
                edge_dict_of_dicts = defaultdict[Any, Dict[Any, Any]](dict)
                for (source, target), edge_dict in dict(graph.edges).items():
                    edge_dict_of_dicts[source][target] = edge_dict
                return {'nodes': to_state_dict(dict(graph.nodes)),
                        'edges': to_state_dict(dict(edge_dict_of_dicts))}

            def ty_from_state_dict(graph: T, state_dict: Dict[str, Any]) -> T:
                retval = type(graph)()
                for node_name, node_dict in state_dict['nodes'].items():
                    retval.add_node(node_name, **from_state_dict(graph.nodes[node_name], node_dict))
                for source, target_and_edge_dict in state_dict['edges'].items():
                    for target, edge_dict in target_and_edge_dict.items():
                        retval.add_edge(source, target,
                                        **from_state_dict(graph.edges[source, target], edge_dict))
                return retval

            register_serialization_state(cls, ty_to_state_dict,  # type: ignore[no-untyped-call]
                                         ty_from_state_dict)

        register_graph_as_flax_state_dict(nx.Graph)
        register_graph_as_flax_state_dict(nx.DiGraph)

    @display_generic.register
    def _(value: nx.Graph,
          seen: Set[int],
          show_values: bool = True,
          indent: int = 0,
          batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
        directed = isinstance(value, nx.DiGraph)
        arrow = cf.base00('⟶  ' if directed else '🡘 ')
        retval = display_class(type(value))
        bdi = BatchDimensionIterator(batch_dims)
        for name, node in value.nodes.items():
            sub_batch_dims = bdi.advance(node)
            retval += display_key_and_value(name, node, ": ", seen, show_values, indent + 1,
                                            sub_batch_dims)
        for (source, target), edge in value.edges.items():
            key = f"{source}{arrow}{target}"
            sub_batch_dims = bdi.advance(edge)
            retval += display_key_and_value(key, edge, ": ", seen, show_values, indent + 1,
                                            sub_batch_dims)
        return retval



from __future__ import annotations

from functools import partial, singledispatch
from numbers import Complex, Number, Real
from typing import Any, Dict, List, Optional, Tuple, Union

import jax.numpy as jnp
import numpy as np
from jax.interpreters.xla import DeviceArray
from jax.tree_util import tree_flatten, tree_map, tree_reduce

from .annotations import Array, BooleanNumeric, PyTree
from .dtypes import default_tols

__all__ = ['assert_tree_allclose', 'tree_allclose', 'get_test_string', 'get_relative_test_string']


def assert_tree_allclose(actual: PyTree,
                         desired: PyTree,
                         original_name: Optional[str] = None,
                         original_value: Optional[PyTree] = None,
                         *,
                         rtol: Optional[float] = None,
                         atol: Optional[float] = None) -> None:
    """
    Asserts that every tensor in an actual pytree matches the corresponding tensor in a desired
    pytree.  If the assertion fails, a passing test string is printed::

    ```python
    from tjax import assert_tree_allclose, dataclass, RealNumeric

    @dataclass
    class A:
        x: RealNumeric
        y: RealNumeric

    @dataclass
    class B:
        z: A

    original = B(A(1.2, 3.4))
    desired = B(A(3.0, 4.0))
    actual = B(A(1.2, 5.2))

    assert_tree_allclose(actual, desired)
    ```
    This prints::
    ```
    JAX trees don't match with rtol=0.0001 and atol=1e-06.
    Mismatched elements: 2 / 2 (100%)
    Maximum absolute difference: 1.8
    Maximum relative difference: 0.6
    Actual: B
        z=A
            x=1.2
            y=5.2

    Desired: B
        z=A
            x=3.0
            y=4.0

    Test string:
    actual = B(z=A(x=1.2, y=5.2))
    ```
    The test string can then be pasted.

    Args:
        actual: The actual value.
        desired: The desired value.
        original_name: The variable name that contains the original value.
        original_value: The original value.  This is usually a pytree like a dataclass that has the
            same type as actual and desired, but contains different values.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    flattened_actual, structure_actual = tree_flatten(actual)
    flattened_desired, structure_desired = tree_flatten(desired)
    if structure_actual != structure_desired:
        raise AssertionError(f"\nTree structure mismatch.\nActual: {actual}\nDesired: {desired}\n")

    for i, (actual_, desired_) in enumerate(zip(flattened_actual, flattened_desired)):
        dtype = jnp.result_type(actual_, desired_)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)  # pyright: ignore
        try:
            np.testing.assert_allclose(actual_, desired_, **tols)
        except AssertionError as exception:
            old_message = exception.args[0].split('\n')
            best_part_of_old_message = "\n".join(old_message[3:6]).replace("Max ", "Maximum ")
            test_string = (get_relative_test_string(actual, original_name, original_value, **tols)
                           if original_name is not None and original_value is not None
                           else get_test_string(actual, **tols))
            test_string = "desired = " + test_string
            # style_config = yapf.style.CreatePEP8Style()
            # style_config['COLUMN_LIMIT'] = column_limit
            # test_string = yapf.yapf_api.FormatCode(test_string, style_config=style_config)[0]
            message = (
                f"\nTree leafs don't match at position {i} with rtol={tols['rtol']} and "
                f"atol={tols['atol']}.\n"
                f"{best_part_of_old_message}\n\n"
                f"Actual: {actual}\nDesired: {desired}\n"
                f"Test string:\n{test_string}")
            raise AssertionError(message) from None


def tree_allclose(actual: PyTree,
                  desired: PyTree,
                  rtol: Optional[float] = None,
                  atol: Optional[float] = None) -> bool:
    """
    Args:
        actual: The actual value.
        desired: The desired value.
        rtol: The relative tolerance of the comparisons in the comparison.
        atol: The absolute tolerance of the comparisons in the comparison.
    """
    def allclose(actual_array: Array, desired_array: Array) -> BooleanNumeric:
        dtype = jnp.result_type(actual_array, desired_array)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)  # pyright: ignore
        return bool(jnp.allclose(actual_array, desired_array, **tols))

    return tree_reduce(jnp.logical_and, tree_map(allclose, actual, desired), True)


# get test string ----------------------------------------------------------------------------------
# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.
@singledispatch
def get_test_string(actual: Any, rtol: float, atol: float) -> str:
    """
    Args:
        actual: The actual value that was produced, and that should be the desired value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    Returns:
        A string of Python code that produces the desired value.
    """
    return str(actual)


@get_test_string.register(np.ndarray)
@get_test_string.register(DeviceArray)
def _(actual: Union[Array, DeviceArray], rtol: float, atol: float) -> str:
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ')


@get_test_string.register
def _(actual: Complex, rtol: float, atol: float) -> str:
    x = float(actual) if isinstance(actual, Real) else complex(actual)
    return _inexact_number_to_string(x, rtol, atol)


@get_test_string.register
def _(actual: Number, rtol: float, atol: float) -> str:
    return str(actual)


@get_test_string.register(list)
@get_test_string.register(tuple)
def _(actual: Union[List[Any], Tuple[Any]], rtol: float, atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_test_string(sub_actual, rtol, atol)
                        for i, sub_actual in enumerate(actual))
            + ("]" if is_list else ")"))


@get_test_string.register(dict)
def _(actual: Dict[Any, Any], rtol: float, atol: float) -> str:
    return '{' + ",\n".join(repr(key) + ': ' + get_test_string(sub_actual, rtol, atol)
                            for key, sub_actual in actual.items()) + '}'


# get relative test string -------------------------------------------------------------------------
@singledispatch
def get_relative_test_string(actual: Any,
                             original_name: str,
                             original: Any,
                             rtol: float,
                             atol: float) -> str:
    """
    Args:
        original_name: The name of the variable containing an original value.
        actual: The actual value that was produced, and that should be the desired value.
        original: The original value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    Returns:
        A string of Python code that produces the desired value from an "original value" (could be
        zeroed-out, for example).
    """
    return str(actual)


@get_relative_test_string.register(np.ndarray)
@get_relative_test_string.register(DeviceArray)
def _(actual: Union[Array, DeviceArray], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ')


@get_relative_test_string.register
def _(actual: Complex, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return _inexact_number_to_string(actual, rtol, atol)  # type: ignore[arg-type]


@get_relative_test_string.register
def _(actual: Number, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return str(actual)


@get_relative_test_string.register(list)
@get_relative_test_string.register(tuple)
def _(actual: Union[List[Any], Tuple[Any]], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_relative_test_string(f"{original_name}[{i}]",
                                                 sub_actual, sub_original,
                                                 rtol, atol)
                        for i, (sub_actual, sub_original) in enumerate(zip(actual, original)))
            + ("]" if is_list else ")"))


@get_relative_test_string.register(dict)
def _(actual: Dict[Any, Any], original_name: str, original: Any, rtol: float, atol: float) -> str:
    if not isinstance(original, dict):
        raise TypeError

    def relative_string(key: Any, sub_actual: Any) -> str:
        return get_relative_test_string(
            f"{original_name}[{key}]", sub_actual, original[key], rtol, atol)

    return '{' + ",\n".join(repr(key) + ': ' + relative_string(key, sub_actual)
                            for key, sub_actual in actual.items()) + '}'


# Private functions --------------------------------------------------------------------------------
def _float_to_string_with_precision(x: Union[float, complex], precision: int) -> str:
    with np.printoptions(precision=precision, floatmode='maxprec'):
        return repr(np.array(x))[6:-1]


def _inexact_number_to_string(x: Union[complex, np.inexact[Any]], rtol: float, atol: float) -> str:
    y: Union[float, complex]
    if isinstance(x, Real):  # type: ignore[unreachable]
        y = float(x)  # type: ignore[unreachable]
    elif isinstance(x, Complex):  # type: ignore[unreachable]
        y = complex(x)  # type: ignore[unreachable]
    else:
        raise TypeError
    retval = ""  # type: ignore[unreachable]
    for i in range(20):
        retval = _float_to_string_with_precision(y, i)
        if np.allclose(complex(retval), x, rtol=rtol, atol=atol):
            break
    return retval



from __future__ import annotations

from typing import Generic, TypeVar

from .annotations import PyTree, RealArray
from .dataclasses import dataclass

__all__ = ['PlottableTrajectory']


Trajectory = TypeVar('Trajectory', bound=PyTree)


@dataclass
class PlottableTrajectory(Generic[Trajectory]):
    trajectory: Trajectory
    "The trajectory is a PyTree containing the plotted data in its dynamic attributes."
    times: RealArray
    "The times corresponding to the data points in each of the plotted attributes."



from __future__ import annotations

from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from functools import singledispatch
from numbers import Number
from typing import Any, Iterable, List, Optional, Set, Tuple, Type, TypeVar, Union

import colorful as cf
import numpy as np
from jax.errors import TracerArrayConversionError
from jax.experimental.host_callback import id_tap
from jax.interpreters.ad import JVPTracer
from jax.interpreters.batching import BatchTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer, JaxprTracer
from jax.interpreters.xla import DeviceArray
from jax.tree_util import tree_leaves

from .annotations import Array, TapFunctionTransforms

__all__ = ['print_generic', 'display_generic', 'id_display']


FlaxModule: Type[Any]
try:
    from flax.linen import Module as FlaxModule  # type: ignore[import]
    flax_loaded = True
except ImportError:
    flax_loaded = False
    FlaxModule = type(None)


# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/11727.


def print_generic(*args: Any,
                  batch_dims: Optional[Tuple[Optional[int], ...]] = None,
                  raise_on_nan: bool = True,
                  **kwargs: Any) -> None:
    bdi = BatchDimensionIterator(batch_dims)
    found_nan = False
    for value in args:
        sub_batch_dims = bdi.advance(value)
        s = display_generic(value, set(), batch_dims=sub_batch_dims)
        print(s)
        found_nan = found_nan or raise_on_nan and 'nan' in str(s)
    for key, value in kwargs.items():
        sub_batch_dims = bdi.advance(value)
        s = display_key_and_value(key, value, "=", set(), batch_dims=sub_batch_dims)
        print(s)
        found_nan = found_nan or raise_on_nan and 'nan' in str(s)
    if found_nan:
        assert False


@singledispatch
def display_generic(value: Any,
                    seen: Set[int],
                    show_values: bool = True,
                    indent: int = 0,
                    batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    if is_dataclass(value) and not isinstance(value, type):
        return display_dataclass(value, seen, show_values, indent, batch_dims)
    return cf.red(str(value)) + "\n"


@display_generic.register(type)
def _(value: Type[Any],
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.red(f"type[{value.__name__}]") + "\n"


@display_generic.register
def _(value: JVPTracer,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"JVPTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register
def _(value: JaxprTracer,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"JaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register
def _(value: DynamicJaxprTracer,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"DynamicJaxprTracer {value.shape} {value.dtype}") + "\n"


@display_generic.register
def _(value: BatchTracer,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.magenta(f"BatchTracer {value.shape} {value.dtype} "
                      f"batched over {value.val.shape[value.batch_dim]}") + "\n"


@display_generic.register(np.ndarray)
def _(value: Array,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    retval = cf.yellow(f"NumPy Array {value.shape} {value.dtype}{_batch_str(batch_dims)}") + "\n"
    return retval + _show_array(indent + 1, value)


@display_generic.register
def _(value: DeviceArray,
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    base_string = cf.violet(
        f"Jax Array {value.shape} {value.dtype}") + "\n"  # type: ignore[attr-defined]
    try:
        np_value = np.asarray(value)
    except TracerArrayConversionError:
        array_string = ""
    else:
        array_string = _show_array(indent + 1, np_value)
    return base_string + array_string


@display_generic.register(type(None))
@display_generic.register(Number)
def _(value: Union[None, Number],
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return cf.cyan(str(value)) + "\n"


@display_generic.register(Mapping)
def _(value: Mapping[Any, Any],
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return (display_class(type(value))
            + "".join(display_key_and_value(key, sub_value, "=", seen, show_values, indent + 1)
                      for key, sub_value in sorted(value.items())))


@display_generic.register(tuple)
@display_generic.register(list)
def _(value: Union[Tuple[Any, ...], List[Any]],
      seen: Set[int],
      show_values: bool = True,
      indent: int = 0,
      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    return display_class(type(value)) + "".join(
        display_key_and_value("", sub_value, "", seen, show_values, indent + 1, sub_batch_dims)
        for sub_batch_dims, sub_value in zip(batch_dimension_iterator(value, batch_dims), value))


def display_dataclass(value: Any,
                      seen: Set[int],
                      show_values: bool = True,
                      indent: int = 0,
                      batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    is_module = flax_loaded and isinstance(value, FlaxModule)
    retval = display_class(type(value), is_module)
    bdi = BatchDimensionIterator(batch_dims)
    for field_info in fields(value):
        name = field_info.name
        if is_module and name in {'parent', 'name'}:
            continue
        sub_value = getattr(value, name)
        sub_batch_dims = bdi.advance(sub_value)
        retval += display_key_and_value(name, sub_value, "=", seen, show_values, indent + 1,
                                        sub_batch_dims)
    if is_module:
        retval += display_key_and_value('name', value.name, "=", seen, show_values, indent + 1,
                                        None)
        retval += display_key_and_value('has_parent', value.parent is not None, "=", seen,
                                        show_values, indent + 1, None)
        retval += display_key_and_value('bound', value.scope is not None, "=", seen, show_values,
                                        indent + 1, None)
        # pylint: disable=protected-access
        for name, child_module in value._state.children.items():  # pytype: disable=attribute-error
            if not isinstance(child_module, FlaxModule):
                continue
            retval += display_key_and_value(name, child_module, "=", seen, show_values, indent + 1,
                                            None)
    return retval


_T = TypeVar('_T')


def id_display(x: _T, name: Optional[str] = None, *, no_jvp: bool = False) -> _T:
    def tap(x: _T, transforms: TapFunctionTransforms) -> None:
        nonlocal name
        batch_dims: Optional[Tuple[Optional[int], ...]] = None
        flags = []
        for transform_name, transform_dict in transforms:
            if transform_name == 'batch':
                batch_dims = transform_dict['batch_dims']
                continue
            if no_jvp and transform_name == 'jvp':
                return
            if transform_name in ['jvp', 'mask', 'transpose']:
                flags.append(transform_name)
                continue
        if name is None:
            print_generic(x, batch_dims=batch_dims)
        else:
            if flags:
                final_name = name + f" [{', '.join(flags)}]"
            else:
                final_name = name
            # https://github.com/python/mypy/issues/11583
            print_generic(batch_dims=batch_dims, **{final_name: x})  # type: ignore[arg-type]
    return id_tap(tap, x, result=x)  # type: ignore[no-untyped-call]


class BatchDimensionIterator:
    def __init__(self, batch_dims: Optional[Tuple[Optional[int], ...]] = None):
        super().__init__()
        self.batch_dims = batch_dims
        self.i = 0

    def advance(self, value: Any) -> Optional[Tuple[Optional[int], ...]]:
        if self.batch_dims is None:
            return None
        n = len(tree_leaves(value))
        old_i = self.i
        self.i += n
        return self.batch_dims[old_i: self.i]

    def check_done(self) -> None:
        if self.batch_dims is not None:
            assert self.i == len(self.batch_dims)


def batch_dimension_iterator(values: Iterable[Any],
                             batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> Iterable[
                                 Optional[Tuple[Optional[int], ...]]]:
    bdi = BatchDimensionIterator(batch_dims)
    for value in values:
        yield bdi.advance(value)
    bdi.check_done()


# Public unexported functions ----------------------------------------------------------------------
def display_class(cls: Type[Any], is_module: bool = False) -> str:
    color_f = cf.green if is_module else cf.orange
    return color_f(cls.__name__) + "\n"


def display_key_and_value(key: str,
                          value: Any,
                          separator: str,
                          seen: Set[int],
                          show_values: bool = True,
                          indent: int = 0,
                          batch_dims: Optional[Tuple[Optional[int], ...]] = None) -> str:
    if id(value) in seen:
        value = "<seen>"
    elif is_dataclass(value) and not isinstance(value, type):
        seen.add(id(value))
    return (_indent_space(indent) + cf.blue(key) + cf.base00(separator)
            + display_generic(value, seen, show_values, indent, batch_dims))


# Private functions --------------------------------------------------------------------------------
_indentation = 4


def _batch_str(batch_dims: Optional[Tuple[Optional[int], ...]]) -> str:
    if batch_dims is None:
        return ""
    assert len(batch_dims) == 1
    batch_dim = batch_dims[0]
    if batch_dim is None:
        return ""
    return f" batched over axis {batch_dim}"


def _indent_space(indent: int) -> str:
    return (indent * _indentation) * " "


@singledispatch
def _format_number(x: Any) -> str:
    return f"{x:10}"


@_format_number.register(np.inexact)
def _(x: np.inexact[Any]) -> str:
    return f"{x:10.4f}"


def _show_array(indent: int, array: Array) -> str:
    if not np.issubdtype(array.dtype, np.number) and not np.issubdtype(array.dtype, np.bool_):
        return ""
    if np.prod(array.shape) == 0:
        return ""
    if any(x > 12 for x in array.shape):
        return ""
    if 1 in array.shape:
        return _show_array(indent, array[tuple[Union[int, slice], ...](0 if s == 1 else slice(None)
                                                                       for s in array.shape)])
    if len(array.shape) == 0:
        return _indent_space(indent) + _format_number(array[()]) + "\n"
    if len(array.shape) == 1:
        return _indent_space(indent) + "  ".join(_format_number(array[i])
                                                 for i in range(array.shape[0])) + "\n"
    if len(array.shape) == 2:
        return "".join(_show_array(indent, array_slice)
                       for array_slice in array)
    return ""



from __future__ import annotations

from typing import Callable, Generic, Optional, Tuple, TypeVar

from jax.experimental.host_callback import id_tap
from jax.lax import scan

from ..annotations import PyTree, TapFunctionTransforms
from ..dataclasses import dataclass
from .augmented import State

__all__ = ['IteratedFunctionBase']


Parameters = TypeVar('Parameters', bound=PyTree)
Trajectory = TypeVar('Trajectory', bound=PyTree)
TheAugmentedState = TypeVar('TheAugmentedState')
TapFunction = Callable[[None, TapFunctionTransforms], None]


@dataclass
class IteratedFunctionBase(Generic[Parameters, State, Trajectory, TheAugmentedState]):
    """
    An IteratedFunctionBase object models any iterated function.

    It is a generic class in terms of four generic types, all of which are pytrees:
        * Parameters, which models the iteration parameters,
        * State, which is the state at each iteration,
        * Trajectory, which is the type produced by sample_trajectory, and
        * TheAugmentedState, which is typically defined by one of the tjax subclasses.
    """
    # New methods ----------------------------------------------------------------------------------
    def sample_trajectory(self,
                          theta: Parameters,
                          initial_state: State,
                          maximum_iterations: int,
                          tap_function: Optional[TapFunction]) -> (
                              Tuple[TheAugmentedState, Trajectory]):
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
            maximum_iterations: The number of steps in the trajectory.  Unlike the eponymous member
                variable, this must be static.
        Returns:
            x_star: The augmented state at the fixed point.
            trajectory: A PyTree representing the trajectory of states.
        """
        def f(augmented: TheAugmentedState, x: None) -> Tuple[TheAugmentedState, Trajectory]:
            trajectory: Trajectory
            new_state, trajectory = self.sampled_state_trajectory(theta, augmented)
            new_augmented = self.iterate_augmented(new_state, augmented)
            if tap_function is not None:
                trajectory = id_tap(tap_function,  # type: ignore[no-untyped-call]
                                    None, result=trajectory)
            return new_augmented, trajectory
        return scan(f, self.initial_augmented(initial_state), None, maximum_iterations)

    # Abstract methods -----------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> TheAugmentedState:
        raise NotImplementedError

    def sampled_state(self, theta: Parameters, state: State) -> State:
        """
        Returns: A sampled value of the next state in a trajectory.  This is used when finding the
            fixed point.  It is included in this base class because it is nearly always called by
            sampled_state_trajectory.
        """
        raise NotImplementedError

    def sampled_state_trajectory(self,
                                 theta: Parameters,
                                 augmented: TheAugmentedState) -> Tuple[State, Trajectory]:
        """
        Returns:
            sampled_state: A sampled value of the next state in a trajectory.
            trajectory: A value to be concatenated into a trajectory.
        """
        raise NotImplementedError

    def iterate_augmented(self,
                          new_state: State,
                          augmented: TheAugmentedState) -> TheAugmentedState:
        """
        Args:
            new_state: The new state to fold into the augmented state.
            augmented: The last augmented state.
        Returns: The next augmented state.
        """
        raise NotImplementedError



# pylint: disable=redefined-builtin
from __future__ import annotations

import dataclasses
import sys
from dataclasses import MISSING, fields
from typing import Any, Callable, Dict, Mapping, Optional, TypeVar, overload

__all__ = ['field', 'as_shallow_dict']


T = TypeVar('T', bound=Any)


if sys.version_info >= (3, 10):
    # NOTE: Actual return type is 'Field[T]', but we want to help type checkers
    # to understand the magic that happens at runtime.
    @overload  # `default` and `default_factory` are optional and mutually exclusive.
    def field(*, static: bool = False, default: T, init: bool = ..., repr: bool = ...,
              hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ..., kw_only: bool = ...) -> T:
        ...

    @overload
    def field(*, static: bool = False, default_factory: Callable[[], T], init: bool = ...,
              repr: bool = ..., hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ..., kw_only: bool = ...) -> T:
        ...

    @overload
    def field(*, static: bool = False, init: bool = ..., repr: bool = ...,
              hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ..., kw_only: bool = ...) -> Any:
        ...

    def field(*, static: bool = False, default: Any = MISSING,
              default_factory: Any = MISSING, init: bool = True,
              repr: bool = True, hash: Optional[bool] = None, compare: bool = True,
              metadata: Optional[Mapping[str, Any]] = None,
              kw_only: Any = MISSING) -> Any:
        """
        Args:
            static: Indicates whether a field is a pytree or static.  Pytree fields are
                differentiated and traced.  Static fields are hashed and compared.
        """
        metadata_dict: Dict[str, Any] = {} if metadata is None else dict(metadata)
        metadata_dict['static'] = static
        if default is MISSING:
            return dataclasses.field(default_factory=default_factory,
                                     init=init,
                                     repr=repr,
                                     hash=hash,
                                     compare=compare,
                                     metadata=metadata_dict,
                                     kw_only=kw_only)
        return dataclasses.field(default=default,
                                 init=init,
                                 repr=repr,
                                 hash=hash,
                                 compare=compare,
                                 metadata=metadata_dict,
                                 kw_only=kw_only)
else:
    # NOTE: Actual return type is 'Field[T]', but we want to help type checkers
    # to understand the magic that happens at runtime.
    @overload  # `default` and `default_factory` are optional and mutually exclusive.
    def field(*, static: bool = False, default: T, init: bool = ..., repr: bool = ...,
              hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ...) -> T:
        ...

    @overload
    def field(*, static: bool = False, default_factory: Callable[[], T], init: bool = ...,
              repr: bool = ..., hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ...) -> T:
        ...

    @overload
    def field(*, static: bool = False, init: bool = ..., repr: bool = ...,
              hash: Optional[bool] = ..., compare: bool = ...,
              metadata: Optional[Mapping[str, Any]] = ...) -> Any:
        ...

    def field(*, static: bool = False, default: Any = MISSING,
              default_factory: Callable[[], Any] = MISSING, init: bool = True,
              repr: bool = True, hash: Optional[bool] = None, compare: bool = True,
              metadata: Optional[Mapping[str, Any]] = None) -> Any:
        """
        Args:
            static: Indicates whether a field is a pytree or static.  Pytree fields are
                differentiated and traced.  Static fields are hashed and compared.
        """
        metadata_dict: Dict[str, Any] = {} if metadata is None else dict(metadata)
        metadata_dict['static'] = static
        if default is MISSING:
            return dataclasses.field(default_factory=default_factory,
                                     init=init,
                                     repr=repr,
                                     hash=hash,
                                     compare=compare,
                                     metadata=metadata_dict)
        return dataclasses.field(default=default,
                                 init=init,
                                 repr=repr,
                                 hash=hash,
                                 compare=compare,
                                 metadata=metadata_dict)


def as_shallow_dict(dcls: Any) -> Dict[str, Any]:
    return {field.name: getattr(dcls, field.name) for field in fields(dcls)}



from __future__ import annotations

from functools import partial
from typing import Optional, Tuple, TypeVar

from .display import id_display
from .shims import custom_vjp

__all__ = ['copy_cotangent', 'replace_cotangent', 'print_cotangent']


X = TypeVar('X')


# copy_cotangent -----------------------------------------------------------------------------------
@custom_vjp
def copy_cotangent(x: X, y: X) -> X:
    return x


def _copy_cotangent_fwd(x: X, y: X) -> Tuple[X, None]:
    return x, None


def _copy_cotangent_bwd(residuals: None, x_bar: X) -> Tuple[X, X]:
    del residuals
    return x_bar, x_bar


# Pyright can't infer types because custom_vjp doesn't yet depend on ParamSpec.
copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)


# replace_cotangent --------------------------------------------------------------------------------
@custom_vjp
def replace_cotangent(x: X, new_cotangent: X) -> X:
    return x


def _replace_cotangent_fwd(x: X, new_cotangent: X) -> Tuple[X, X]:
    return x, new_cotangent


def _replace_cotangent_bwd(residuals: X, x_bar: X) -> Tuple[X, X]:
    return residuals, x_bar


# Pyright can't infer types because custom_vjp doesn't yet depend on ParamSpec.
replace_cotangent.defvjp(_replace_cotangent_fwd, _replace_cotangent_bwd)


# print_cotangent ----------------------------------------------------------------------------------
@partial(custom_vjp, static_argnums=(1,))
def print_cotangent(x: X, name: Optional[str] = None) -> X:
    return x


def _print_cotangent_fwd(x: X, name: Optional[str]) -> Tuple[X, None]:
    return x, None


def _print_cotangent_bwd(name: Optional[str], residuals: None, x_bar: X) -> Tuple[X]:
    del residuals
    x_bar = id_display(x_bar, name)
    return (x_bar,)


print_cotangent.defvjp(_print_cotangent_fwd, _print_cotangent_bwd)  # pyright: ignore



from __future__ import annotations

from typing import Any, Optional, Type, TypedDict

import jax.numpy as jnp

__all__ = ['default_rtol',
           'default_atol',
           'default_tols']


class Tols(TypedDict):
    rtol: float
    atol: float


def default_rtol(dtype: Type[jnp.number[Any]]) -> float:
    if not jnp.issubdtype(dtype, jnp.inexact):
        return 0.0
    return {jnp.bfloat16: 1e-4,
            jnp.float16: 1e-3, jnp.float32: 1e-4, jnp.float64: 1e-5,
            jnp.complex64: 1e-4, jnp.complex128: 1e-5}[dtype]  # pyright: ignore


def default_atol(dtype: Type[jnp.number[Any]]) -> float:
    if not jnp.issubdtype(dtype, jnp.inexact):
        return 1.0
    return {jnp.bfloat16: 1e-2,
            jnp.float16: 1e-3, jnp.float32: 1e-6, jnp.float64: 1e-8,
            jnp.complex64: 1e-6, jnp.complex128: 1e-8}[dtype]  # pyright: ignore


def default_tols(dtype: Type[jnp.number[Any]],
                 *,
                 rtol: Optional[float] = None,
                 atol: Optional[float] = None) -> Tols:
    return Tols(rtol=default_rtol(dtype) if rtol is None else rtol,
                atol=default_atol(dtype) if atol is None else atol)



"""
This library implements a variety of tools for the differential programming library
[JAX](https://github.com/google/jax).
"""
from . import dataclasses, fixed_point, gradient
from ._src.annotations import (Array, BooleanArray, BooleanNumeric, Complex, ComplexArray,
                               ComplexNumeric, Integral, IntegralArray, IntegralNumeric, PyTree,
                               Real, RealArray, RealNumeric, Shape, ShapeLike, SliceLike,
                               TapFunctionTransforms)
from ._src.cotangent_tools import copy_cotangent, print_cotangent, replace_cotangent
from ._src.display import display_generic, id_display, print_generic
from ._src.dtypes import default_atol, default_rtol, default_tols
from ._src.generator import Generator
from ._src.leaky_integral import (diffused_leaky_integrate, leaky_covariance, leaky_data_weight,
                                  leaky_integrate, leaky_integrate_time_series)
from ._src.partial import Partial
from ._src.plottable_trajectory import PlottableTrajectory
from ._src.shims import custom_jvp, custom_vjp, jit
from ._src.testing import (assert_tree_allclose, get_relative_test_string, get_test_string,
                           tree_allclose)
from ._src.tools import abs_square, divide_nonnegative, divide_where, is_scalar, zero_tangent_like

__all__ = ['Array', 'BooleanArray', 'BooleanNumeric', 'Complex', 'ComplexArray', 'ComplexNumeric',
           'Generator', 'Integral', 'IntegralArray', 'IntegralNumeric', 'Partial',
           'PlottableTrajectory', 'PyTree', 'Real', 'RealArray', 'RealNumeric', 'Shape',
           'ShapeLike', 'SliceLike', 'TapFunctionTransforms', 'abs_square', 'assert_tree_allclose',
           'copy_cotangent', 'custom_jvp', 'custom_vjp', 'dataclasses', 'default_atol',
           'default_rtol', 'default_tols', 'diffused_leaky_integrate', 'display_generic',
           'divide_nonnegative', 'divide_where', 'fixed_point', 'get_relative_test_string',
           'get_test_string', 'gradient', 'id_display', 'is_scalar', 'jit', 'leaky_covariance',
           'leaky_data_weight', 'leaky_integrate', 'leaky_integrate_time_series', 'print_cotangent',
           'print_generic', 'replace_cotangent', 'tree_allclose', 'zero_tangent_like']
#
# __pdoc__ = {}
# __pdoc__['PyTreeLike'] = False
# __pdoc__['Field'] = False
# __pdoc__['InitVar'] = False
# __pdoc__['FrozenInstanceError'] = False
#
# document_dataclass(__pdoc__, 'Generator')
# document_dataclass(__pdoc__, 'Partial')
# del document_dataclass



from __future__ import annotations

from typing import Generic

from ..dataclasses import dataclass
from .augmented import State
from .base import IteratedFunctionBase, Parameters, Trajectory

__all__ = ['SimpleScan']


@dataclass
class SimpleScan(IteratedFunctionBase[Parameters, State, Trajectory, State],
                 Generic[Parameters, State, Trajectory]):
    """
    A SimpleScan object models an iterated function that runs for a fixed number of steps.

    It is a generic class in terms of three generic types, all of which are pytrees:
        * Parameters, which models the iteration parameters,
        * State, which is the state at each iteration,
        * Trajectory, which is the type produced by sample_trajectory, and

    The main method of SimpleScan is sample_trajectory, which iterates for a fixed number of
    iterations and returns a trajectory.
    """
    # Implemented methods --------------------------------------------------------------------------
    def initial_augmented(self, initial_state: State) -> State:
        return initial_state

    def iterate_augmented(self,
                          new_state: State,
                          augmented: State) -> State:
        return new_state



from typing import Type

import mypy.plugins.dataclasses
from mypy.plugin import Plugin  # pylint: disable=no-name-in-module


try:
    import flax  # type: ignore[import]  # noqa: F401  # pylint: disable=unused-import
except ImportError:
    has_flax = True
else:
    has_flax = False


class CustomPlugin(Plugin):
    pass


def plugin(version: str) -> Type[CustomPlugin]:
    mypy.plugins.dataclasses.dataclass_makers.add('tjax._src.dataclasses.dataclass.dataclass')
    mypy.plugins.dataclasses.field_makers.add('tjax._src.dataclasses.helpers.field')
    if has_flax:
        mypy.plugins.dataclasses.dataclass_makers.add('flax.struct.dataclasses.dataclass')
        mypy.plugins.dataclasses.field_makers.add('flax.struct.dataclasses.field')
    return CustomPlugin



from __future__ import annotations

from functools import partial
from typing import Any, Generic, Tuple, TypeVar

import jax.numpy as jnp
from jax.lax import while_loop

from ..annotations import BooleanNumeric, IntegralNumeric, PyTree, RealNumeric
from ..dataclasses import dataclass
from .augmented import AugmentedState, State
from .base import IteratedFunctionBase, Parameters, Trajectory

__all__ = ['IteratedFunction']


Comparand = TypeVar('Comparand', bound=PyTree)
TheAugmentedState = TypeVar('TheAugmentedState', bound=AugmentedState[Any])


@dataclass
class IteratedFunction(IteratedFunctionBase[Parameters, State, Trajectory, TheAugmentedState],
                       Generic[Parameters, State, Comparand, Trajectory, TheAugmentedState]):
    """
    An IteratedFunction object models an iterated function.

    It is a generic class in terms of five generic types, all of which are pytrees:
        * Parameters, which models the iteration parameters,
        * State, which is the state at each iteration,
        * Comparand, which is calculated from state and compared between iterations to check
          convergence,
        * Trajectory, which is the type produced by sample_trajectory, and
        * TheAugmentedState, which is typically defined by one of the tjax subclasses.

    The two main methods of IteratedFunction are
    * find_fixed_point, which iterates until the fixed point is reached or other stopping conditions
      are met, and
    * sample_trajectory, which iterates for a fixed number of iterations and returns a trajectory.

    Args:
        minimum_iterations: The minimum number of iterations for this fixed point solver.  This must
            be positive.
        maximum_iterations: The maximum number of iterations for this fixed point solver.  This must
            be positive.
        rtol: The relative tolerance for the comparison stopping condition.
        atol: The absolute tolerance for the comparison stopping condition.
    """
    minimum_iterations: IntegralNumeric
    maximum_iterations: IntegralNumeric
    rtol: RealNumeric
    atol: RealNumeric

    # New methods ----------------------------------------------------------------------------------
    def find_fixed_point(self,
                         theta: Parameters,
                         initial_state: State) -> TheAugmentedState:
        """
        Args:
            theta: The parameters for which gradients can be calculated.
            initial_state: An initial guess of the final state.
        Returns:
            state: The augmented state at the fixed point.
        """
        def f(augmented: TheAugmentedState) -> TheAugmentedState:
            new_state = self.sampled_state(theta, augmented.current_state)
            return self.iterate_augmented(new_state, augmented)

        return while_loop(partial(self.state_needs_iteration, theta),
                          f,
                          self.initial_augmented(initial_state))

    def debug_fixed_point(self, theta: Parameters, initial_state: State) -> TheAugmentedState:
        """
        This method is identical to find_fixed_point, but avoids using while_loop.
        """
        augmented = self.initial_augmented(initial_state)
        while self.state_needs_iteration(theta, augmented):
            new_state = self.sampled_state(theta, augmented.current_state)
            augmented = self.iterate_augmented(new_state, augmented)
        return augmented

    def state_needs_iteration(self, theta: Parameters, augmented: TheAugmentedState) -> bool:
        """
        Args:
            theta: The parameters.
            augmented: The state.
        Returns: True while iteration needs to continue.
        """
        enough_iterations = augmented.iterations >= self.minimum_iterations
        converged = self.converged(augmented)
        not_too_many_iterations = augmented.iterations < self.maximum_iterations
        return jnp.logical_and(not_too_many_iterations,
                               jnp.logical_not(jnp.logical_and(enough_iterations, converged)))

    # Abstract methods -----------------------------------------------------------------------------
    def expected_state(self, theta: Parameters, state: State) -> State:
        """
        Returns: The expected value of the next state given the old one.  This is used by the
            combinator.
        """
        raise NotImplementedError

    def converged(self, augmented: TheAugmentedState) -> BooleanNumeric:
        """
        Returns: A Boolean Array of shape () indicating whether the pytrees are close.
        """
        raise NotImplementedError

    def extract_comparand(self, state: State) -> Comparand:
        """
        Returns: The "comparand", which is a subset of the values in the state that are compared by
            subclasses to detect convergence.  In ComparingIteratedFunction, the comparand will be
            compared in successive states to detect convergence.  In StochasticIteratedFunction, the
            mean and variance of the comparand are used to detect convergence.
        """
        raise NotImplementedError

    def minimum_tolerances(self, augmented: TheAugmentedState) -> Tuple[RealNumeric, RealNumeric]:
        """
        Returns:
            The minimum value of atol that would lead to convergence now.
            The minimum value of rtol that would lead to convergence now.
        """
        raise NotImplementedError



from __future__ import annotations

from typing import Any, Generic, List, Optional, Tuple

from ..annotations import PyTree
from ..dataclasses import dataclass
from .transform import GradientState, GradientTransformation, Weights

__all__ = ['ChainedGradientTransformation']


@dataclass
class ChainedGradientState(GradientState):
    sub_states: List[PyTree]


@dataclass
class ChainedGradientTransformation(GradientTransformation[ChainedGradientState, Weights],
                                    Generic[Weights]):
    transforms: List[GradientTransformation[Any, Weights]]

    def init(self, parameters: Weights) -> ChainedGradientState:
        return ChainedGradientState([transform.init(parameters)
                                     for transform in self.transforms])

    def update(self,
               gradient: Weights,
               state: ChainedGradientState,
               parameters: Optional[Weights]) -> Tuple[Weights, ChainedGradientState]:
        new_state: List[PyTree] = []
        for sub_state, transform in zip(state.sub_states, self.transforms):
            gradient, new_sub_state = transform.update(gradient, sub_state, parameters)
            new_state.append(new_sub_state)
        return gradient, ChainedGradientState(new_state)



from tjax._src.gradient.aliases import (DPSGD, LARS, SGD, SM3, AdaBelief, AdaFactor, AdaGrad, Adam,
                                        AdamW, Fromage, Lamb, NoisySGD, RAdam, RMSProp, Yogi)
from tjax._src.gradient.chain import ChainedGradientTransformation
from tjax._src.gradient.smd import SMDGradient, SMDState
from tjax._src.gradient.transform import (GradientState, GradientTransformation,
                                          SecondOrderGradientTransformation,
                                          ThirdOrderGradientTransformation)
from tjax._src.gradient.transforms import (AddDecayedWeights, AddNoise, ApplyEvery, Centralize, Ema,
                                           Scale, ScaleByAdam, ScaleByBelief, ScaleByParamBlockNorm,
                                           ScaleByParamBlockRMS, ScaleByRAdam, ScaleByRms,
                                           ScaleByRss, ScaleBySchedule, ScaleBySM3, ScaleByStddev,
                                           ScaleByTrustRatio, ScaleByYogi, Schedule, Trace)

__all__ = ['ChainedGradientTransformation', 'SMDGradient', 'SMDState',
           # transform.py
           'GradientState', 'GradientTransformation', 'SecondOrderGradientTransformation',
           'ThirdOrderGradientTransformation',
           # aliases.py
           'AdaBelief', 'AdaFactor', 'AdaGrad', 'Adam', 'AdamW', 'Fromage', 'LARS', 'Lamb',
           'NoisySGD', 'RAdam', 'RMSProp', 'SGD', 'SM3', 'Yogi', 'DPSGD',
           # transforms.py
           'Trace', 'Ema', 'ScaleByRss', 'ScaleByRms', 'ScaleByStddev', 'ScaleByAdam', 'Scale',
           'ScaleByParamBlockNorm', 'ScaleByParamBlockRMS', 'ScaleByBelief', 'ScaleByYogi',
           'ScaleByRAdam', 'AddDecayedWeights', 'ScaleBySchedule', 'ScaleByTrustRatio', 'AddNoise',
           'ApplyEvery', 'Centralize', 'ScaleBySM3',
           'Schedule']



from __future__ import annotations

from typing import Any, Callable, Concatenate, Generic, Tuple, TypeVar, overload

import jax
from jax.tree_util import Partial
from typing_extensions import ParamSpec

__all__ = ['jit', 'custom_jvp', 'custom_vjp']


R_co = TypeVar('R_co', covariant=True)
F = TypeVar('F', bound=Callable[..., Any])
P = ParamSpec('P')
T = TypeVar("T", bound="custom_vjp[Any, Any, Any]")
V = TypeVar("V", bound="custom_jvp[Any, Any, Any]")
U = TypeVar("U")


def jit(func: F, **kwargs: Any) -> F:
    """
    This version of jit ensures that abstract methods stay abstract.
    """
    retval = jax.jit(func, **kwargs)
    if hasattr(func, "__isabstractmethod__"):
        retval.__isabstractmethod__ = func.__isabstractmethod__  # type: ignore[attr-defined]
    # Fixed by https://github.com/NeilGirdhar/jax/tree/jit_annotation.
    return retval  # type: ignore[return-value]


class custom_vjp(Generic[U, P, R_co]):
    """
    This is a shim class over jax.custom_vjp to:

    - allow custom_vjp to be used on methods, and
    - rename nondiff_argnums to static_argnums.
    """
    vjp: jax.custom_vjp[R_co]

    def __init__(self,
                 fun: Callable[Concatenate[U, P], R_co],
                 static_argnums: Tuple[int, ...] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
        """
        super().__init__()
        static_argnums = tuple(sorted(static_argnums))
        self.vjp = jax.custom_vjp(fun, nondiff_argnums=static_argnums)

    def defvjp(self,
               fwd: Callable[Concatenate[U, P], Tuple[R_co, Any]],
               bwd: Callable[..., Any]) -> None:
        self.vjp.defvjp(fwd, bwd)

    def __call__(self,
                 u: U,
                 /,
                 *args: P.args,
                 **kwargs: P.kwargs) -> R_co:
        return self.vjp(u, *args, **kwargs)

    @overload
    def __get__(self: T, instance: None, owner: Any = None) -> T:
        ...

    @overload
    def __get__(self, instance: U, owner: Any = None) -> Callable[P, R_co]:
        ...

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R_co]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)  # type: ignore[no-untyped-call]


class custom_jvp(Generic[U, P, R_co]):
    """
    This is a shim class over jax.custom_jvp to:

    - allow custom_vjp to be used on methods, and
    - rename nondiff_argnums to static_argnums.
    """
    def __init__(self,
                 fun: Callable[Concatenate[U, P], R_co],
                 static_argnums: Tuple[int, ...] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
        """
        super().__init__()
        static_argnums = tuple(sorted(static_argnums))
        self.jvp = jax.custom_jvp(fun, nondiff_argnums=static_argnums)

    def defjvp(self, jvp: Callable[Concatenate[U, P], Tuple[R_co, R_co]]) -> None:
        """
        Implement the custom forward pass of the custom derivative.

        Args:
            fwd: The custom forward pass.
        """
        self.jvp.defjvp(jvp)

    def __call__(self, u: U, /, *args: P.args, **kwargs: P.kwargs) -> R_co:
        return self.jvp(u, *args, **kwargs)

    @overload
    def __get__(self: V, instance: None, owner: Any = None) -> V:
        ...

    @overload
    def __get__(self, instance: U, owner: Any = None) -> Callable[P, R_co]:
        ...

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R_co]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)  # type: ignore[no-untyped-call]
